/* methods for dynamically sized array, pool, & stack objects */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "array.h"
#include "datum.h"
#include "defines.h"
#include "simpledefs.h"
#include "prototypes.p"
#include "alloc.h"

#include "array.proto.h"
#include "xprintf.proto.h"

#include "house.proto.h"


char *
SizeArray(struct sarray* array, long size)
{
	array->maxrecs = size;
	return array->data = 
		Reallocate("SizeArray", array->recsize * array->maxrecs, array->data);
}

char *
SizePool(struct sarray* arrayp, long size)
{
	char *ptr;
	long i;
	PoolPtr array;
	array = (PoolPtr)arrayp;
	array->numrecs = array->maxrecs = size;
	array->data = 
		Reallocate("SizePool", array->recsize * array->maxrecs, array->data);
	ptr = (char*)array->data + array->firstfree * array->recsize;
	for (i=array->firstfree; i<array->maxrecs-1; ++i, ptr += array->recsize) {
		((FreeListNodePtr)ptr)->nextfree = i+1;
	}
	((FreeListNodePtr)ptr)->nextfree = -1;
	return array->data; 
}


ArrayPtr 
NewArray(long recsize, long maxrecs)
{
	ArrayPtr array;
	array = (ArrayPtr)MemAlloc("NewArray", (long)sizeof(Array));
	InitArray(array, recsize, maxrecs);
	return array;
}

void
InitArray(ArrayPtr array, long recsize, long maxrecs)
{
	array->numrecs = 0;
	array->maxrecs = maxrecs;
	array->recsize = recsize;
	array->data = MemAlloc("InitArray", maxrecs * recsize);
	array->resize = SizeArray;
}

PoolPtr 
NewPool(long recsize, long maxrecs)
{
	PoolPtr array;
	long i;
	char *ptr;
	array = (PoolPtr)MemAlloc("NewPool", (long)sizeof(Pool));
	InitPool(array, recsize, maxrecs);
	return array;
}


void
InitPool(PoolPtr array, long recsize, long maxrecs)
{
	char *ptr;
	long i;
	InitArray((ArrayPtr)array, recsize, maxrecs);
	array->numrecs = maxrecs;
	FreePool(array);	/* init free list */
	array->resize = SizePool;
	array->numused = 0;
}


char *
GetRecordPtr(ArrayPtr array, long index)
{
	if (index >= array->numrecs) return NULL;
	return (char *)((char*)array->data + index * array->recsize);
}

int
GetRecord(ArrayPtr array, long index, char *record)
{
	if (index >= array->numrecs) return 0;
	memcpy(record, (char*)array->data + index * array->recsize, array->recsize);
	return 1;
}

int
GetNRecords(ArrayPtr array, long index, long n, char *record)
{
	if (index+n >= array->numrecs) return 0;
	memcpy(record, (char*)array->data + index * array->recsize, n * array->recsize);
	return 1;
}

int
PutRecord(ArrayPtr array, long index, char *record)
{
	if (index >= array->numrecs) return 0;
	memcpy((char*)array->data + index * array->recsize, record, array->recsize);
	return 1;
}

int
PutNRecords(ArrayPtr array, long index, int n, char *record)
{
	if (index+n >= array->numrecs) return 0;
	memcpy((char*)array->data + index * array->recsize, record, array->recsize * n);
	return 1;
}

int
InsertRecord(ArrayPtr array, long index, char *record)
{
	char *ptr;
	if (index >= array->numrecs) return 0;
	if (++array->numrecs > array->maxrecs) 
		(*array->resize)(array, array->maxrecs + array->maxrecs);
	ptr = (char*)array->data + index * array->recsize;
	memorymove(ptr + array->recsize, ptr, array->recsize);
	memcpy(ptr, record, array->recsize);
	return 1;
}

int
InsertNRecords(ArrayPtr array, long index, long n, char *record)
{
	char *ptr;
	if (index >= array->numrecs) return 0;
	if ((array->numrecs+=n) >= array->maxrecs)
		(*array->resize)(array, array->maxrecs + n);
	ptr = (char*)array->data + index * array->recsize;
	memorymove(ptr + array->recsize, ptr, array->recsize);
	memcpy(ptr, record, array->recsize);
	return 1;
}

int
DeleteRecord(ArrayPtr array, long index, char *record)
{
	char *ptr;
	-- array->numrecs;
	if (index >= array->numrecs) return 0;
	ptr = (char*)array->data + index * array->recsize;
	if (record != NULL) memcpy(record, ptr, array->recsize);
	memorymove(ptr, ptr + array->recsize, array->recsize);
	return 1;
}

int
DeleteNRecords(ArrayPtr array, long index, long n, char *record)
{
	char *ptr;
	long nbytes;
	if (index+n >= array->numrecs) return 0;
	array->numrecs -= n;
	ptr = (char*)array->data + index * array->recsize;
	nbytes = n * array->recsize;
	if (record != NULL) memcpy(record, ptr, nbytes);
	memorymove(ptr, ptr + nbytes, nbytes);
	return 1;
}

long
AddRecords(ArrayPtr array, long n)
{
	long i;
	i = array->numrecs;
	array->numrecs += n ;
	if (array->numrecs > array->maxrecs) {
		(*array->resize)(array, array->maxrecs + n);
	}
	return i;
}

long
SubRecords(ArrayPtr array, long n)
{
	long i;
	if (array->numrecs-n < 0) return -1;
	array->numrecs -= n ;
	return array->numrecs;
}

int
PushRecord(ArrayPtr array, char *record)
{
	char *ptr;
	if (array->numrecs+1 > array->maxrecs) {
		(*array->resize)(array, array->maxrecs + array->maxrecs);
	}
	ptr = (char*)array->data + array->numrecs * array->recsize;
	memcpy(ptr, record, array->recsize);
	array->numrecs ++ ;
	return 1;
}

long
PushStack(ArrayPtr array)
{
	char *ptr;
	if (array->numrecs+1 > array->maxrecs) {
		(*array->resize)(array, array->maxrecs + array->maxrecs);
	}
	return array->numrecs ++ ;
}

int
PopRecord(ArrayPtr array, char *record)
{
	char *ptr;
	if (array->numrecs == 0) return 0;
	array->numrecs -- ;
	ptr = (char*)array->data + array->numrecs * array->recsize;
	memcpy(record, ptr, array->recsize);
	return 1;
}

int
DropNStack(ArrayPtr array, long n)
{
	if (array->numrecs-n < 0) return 0;
	array->numrecs -= n ;
	return 1;
}


void
FreeRecord(PoolPtr array, long index)
{
	char *ptr;
	ptr = (char*)array->data + index * array->recsize;
	((FreeListNodePtr)ptr)->nextfree = array->firstfree;
	array->firstfree = index;
	array->numused --;
}

void
FreePool(PoolPtr array)
{
	char *ptr;
	long i;
	
	/* initialize free list == free all records */
	array->firstfree = 0;
	ptr = (char*)array->data;
	for (i=0; i<array->maxrecs-1;  i++, ptr = (char *)((char*)ptr + array->recsize)) {
		((FreeListNodePtr)ptr)->nextfree = i+1;
	}
	((FreeListNodePtr)ptr)->nextfree = -1; /* end of list */
}

void
DumpPool(FILE *file, PoolPtr array)
{
	char *ptr;
	long i, j;
	
	/* count free records */
	i = array->firstfree;
	ptr = (char*)((char*)array->data + array->recsize * i);
	j = 0;
	while (i > -1) {
		++ j;
		i = ((FreeListNodePtr)ptr)->nextfree;
		ptr = (char *)((char*)array->data + array->recsize * i);
	}
	xprintf(file, "maxrecs = %ld\n", array->maxrecs);
	xprintf(file, "numused = %ld\n", array->numused);
	xprintf(file, "numfree = %ld\n", j);
	xprintf(file, "unaccounted = %ld\n\n", array->maxrecs-array->numused-j);

}

long
GetFreeRecord(PoolPtr array)
{
	char *ptr;
	long i;
	if (array->firstfree == -1) {
		array->firstfree = array->maxrecs;
		(*array->resize)((struct sarray*)array, array->maxrecs + array->maxrecs);
	}
	i = array->firstfree;
	ptr = (char*)array->data + array->firstfree * array->recsize;
	array->firstfree = ((FreeListNodePtr)ptr)->nextfree;
	array->numused ++;
	return i;
}

long
CopyRecord(PoolPtr array, long index)
{
	register char *ptr1, *ptr2;
	register long newindex;

	newindex = GetFreeRecord(array);
	ptr1 = (char*)array->data + index * array->recsize;
	ptr2 = (char*)array->data + newindex * array->recsize;
	memcpy(ptr2, ptr1, array->recsize);
	return newindex;
}


void
memorymove(char *s1, char *s2, long len)
{
	if (s1 > s2) {
		s1 += len-1;
		s2 += len-1;
		while (len--) *s1-- = *s2--;
	} else {
		while (len--) *s1++ = *s2++;
	}
}


